@@ -296,27 +296,37 @@ event_handle_configurerequest(xcb_configure_request_event_t *ev)
     if((c = client_getbywin(ev->window)))
     {
         area_t geometry = c->geometry;
+        uint16_t bw = c->border_width;
+        uint16_t tb_left = c->titlebar[CLIENT_TITLEBAR_LEFT].size;
+        uint16_t tb_right = c->titlebar[CLIENT_TITLEBAR_RIGHT].size;
+        uint16_t tb_top = c->titlebar[CLIENT_TITLEBAR_TOP].size;
+        uint16_t tb_bottom = c->titlebar[CLIENT_TITLEBAR_BOTTOM].size;
+        uint16_t deco_left = bw + tb_left;
+        uint16_t deco_right = bw + tb_right;
+        uint16_t deco_top = bw + tb_top;
+        uint16_t deco_bottom = bw + tb_bottom;
         int16_t diff_w = 0, diff_h = 0, diff_border = 0;
 
         if(ev->value_mask & XCB_CONFIG_WINDOW_X)
         {
+            int16_t diff = 0;
             geometry.x = ev->x;
-            /* The ConfigureRequest specifies the position of the outer corner of the client window, we want the frame */
-            geometry.x -= c->border_width;
+            xwindow_translate_for_gravity(c->size_hints.win_gravity, deco_left, 0, deco_right, 0, &diff, NULL);
+            geometry.x += diff;
         }
         if(ev->value_mask & XCB_CONFIG_WINDOW_Y)
         {
+            int16_t diff = 0;
             geometry.y = ev->y;
-            /* The ConfigureRequest specifies the position of the outer corner of the client window, we want the frame */
-            geometry.y -= c->border_width;
+            xwindow_translate_for_gravity(c->size_hints.win_gravity, 0, deco_top, 0, deco_bottom, NULL, &diff);
+            geometry.y += diff;
         }
         if(ev->value_mask & XCB_CONFIG_WINDOW_WIDTH)
         {
             uint16_t old_w = geometry.width;
             geometry.width = ev->width;
             /* The ConfigureRequest specifies the size of the client window, we want the frame */
-            geometry.width += c->titlebar[CLIENT_TITLEBAR_LEFT].size;
-            geometry.width += c->titlebar[CLIENT_TITLEBAR_RIGHT].size;
+            geometry.width += tb_left + tb_right;
             diff_w = geometry.width - old_w;
         }
         if(ev->value_mask & XCB_CONFIG_WINDOW_HEIGHT)
@@ -324,13 +334,12 @@ event_handle_configurerequest(xcb_configure_request_event_t *ev)
             uint16_t old_h = geometry.height;
             geometry.height = ev->height;
             /* The ConfigureRequest specifies the size of the client window, we want the frame */
-            geometry.height += c->titlebar[CLIENT_TITLEBAR_TOP].size;
-            geometry.height += c->titlebar[CLIENT_TITLEBAR_BOTTOM].size;
+            geometry.height += tb_top + tb_bottom;
             diff_h = geometry.height - old_h;
         }
         if(ev->value_mask & XCB_CONFIG_WINDOW_BORDER_WIDTH)
         {
-            diff_border = ev->border_width - c->border_width;
+            diff_border = ev->border_width - bw;
             diff_h += diff_border;
             diff_w += diff_border;
 
@@ -392,6 +401,44 @@ event_handle_destroynotify(xcb_destroy_notify_event_t *ev)
             }
 }
 
+/** Record that the given drawable contains the pointer.
+ */
+static void
+event_drawable_under_mouse(lua_State *L, int ud)
+{
+    void *d;
+
+    lua_pushvalue(L, ud);
+    d = luaA_object_ref(L, -1);
+
+    if (d == globalconf.drawable_under_mouse)
+    {
+        /* Nothing to do */
+        luaA_object_unref(L, d);
+        return;
+    }
+
+    if (globalconf.drawable_under_mouse != NULL)
+    {
+        /* Emit leave on previous drawable */
+        luaA_object_push(L, globalconf.drawable_under_mouse);
+        luaA_object_emit_signal(L, -1, "mouse::leave", 0);
+        lua_pop(L, 1);
+
+        /* Unref the previous drawable */
+        luaA_object_unref(L, globalconf.drawable_under_mouse);
+        globalconf.drawable_under_mouse = NULL;
+    }
+    if (d != NULL)
+    {
+        /* Reference the drawable for leave event later */
+        globalconf.drawable_under_mouse = d;
+
+        /* Emit enter */
+        luaA_object_emit_signal(L, ud, "mouse::enter", 0);
+    }
+}
+
 /** The motion notify event handler.
  * \param ev The event.
  */
@@ -419,6 +466,7 @@ event_handle_motionnotify(xcb_motion_notify_event_t *ev)
         if (d)
         {
             luaA_object_push_item(globalconf.L, -1, d);
+            event_drawable_under_mouse(globalconf.L, -1);
             lua_pushnumber(globalconf.L, x);
             lua_pushnumber(globalconf.L, y);
             luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
@@ -431,6 +479,7 @@ event_handle_motionnotify(xcb_motion_notify_event_t *ev)
     {
         luaA_object_push(globalconf.L, w);
         luaA_object_push_item(globalconf.L, -1, w->drawable);
+        event_drawable_under_mouse(globalconf.L, -1);
         lua_pushnumber(globalconf.L, ev->event_x);
         lua_pushnumber(globalconf.L, ev->event_y);
         luaA_object_emit_signal(globalconf.L, -3, "mouse::move", 2);
@@ -444,7 +493,6 @@ event_handle_motionnotify(xcb_motion_notify_event_t *ev)
 static void
 event_handle_leavenotify(xcb_leave_notify_event_t *ev)
 {
-    drawin_t *drawin;
     client_t *c;
 
     globalconf.timestamp = ev->time;
@@ -452,27 +500,18 @@ event_handle_leavenotify(xcb_leave_notify_event_t *ev)
     if(ev->mode != XCB_NOTIFY_MODE_NORMAL)
         return;
 
-    if((c = client_getbyframewin(ev->event)))
+    /* Ignore leave with detail inferior (we were left for a window contained in
+     * our window, so technically the pointer is still inside of this window).
+     */
+    if(ev->detail != XCB_NOTIFY_DETAIL_INFERIOR && (c = client_getbyframewin(ev->event)))
     {
         luaA_object_push(globalconf.L, c);
         luaA_object_emit_signal(globalconf.L, -1, "mouse::leave", 0);
-        drawable_t *d = client_get_drawable(c, ev->event_x, ev->event_y);
-        if (d)
-        {
-            luaA_object_push_item(globalconf.L, -1, d);
-            luaA_object_emit_signal(globalconf.L, -1, "mouse::leave", 0);
-            lua_pop(globalconf.L, 1);
-        }
-        lua_pop(globalconf.L, 1);
     }
 
-    if((drawin = drawin_getbywin(ev->event)))
-    {
-        luaA_object_push(globalconf.L, drawin);
-        luaA_object_push_item(globalconf.L, -1, drawin->drawable);
-        luaA_object_emit_signal(globalconf.L, -1, "mouse::leave", 0);
-        lua_pop(globalconf.L, 2);
-    }
+    lua_pushnil(globalconf.L);
+    event_drawable_under_mouse(globalconf.L, -1);
+    lua_pop(globalconf.L, 1);
 }
 
 /** The enter notify event handler.
@@ -493,23 +532,34 @@ event_handle_enternotify(xcb_enter_notify_event_t *ev)
     {
         luaA_object_push(globalconf.L, drawin);
         luaA_object_push_item(globalconf.L, -1, drawin->drawable);
-        luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
+        event_drawable_under_mouse(globalconf.L, -1);
         lua_pop(globalconf.L, 2);
     }
 
     if((c = client_getbyframewin(ev->event)))
     {
         luaA_object_push(globalconf.L, c);
-        luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
+        /* Ignore enter with detail inferior: The pointer was previously inside
+         * of a child window, so technically this isn't a 'real' enter.
+         */
+        if (ev->detail != XCB_NOTIFY_DETAIL_INFERIOR)
+            luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
+
         drawable_t *d = client_get_drawable(c, ev->event_x, ev->event_y);
         if (d)
         {
             luaA_object_push_item(globalconf.L, -1, d);
-            luaA_object_emit_signal(globalconf.L, -1, "mouse::enter", 0);
+            event_drawable_under_mouse(globalconf.L, -1);
             lua_pop(globalconf.L, 1);
         }
         lua_pop(globalconf.L, 1);
     }
+    else if (ev->event == globalconf.screen->root) {
+        /* When there are multiple X screens with awesome running separate
+         * instances, reset focus.
+         */
+        globalconf.focus.need_update = true;
+    }
 }
 
 /** The focus in event handler.
@@ -594,7 +644,7 @@ event_handle_key(xcb_key_press_event_t *ev)
         /* get keysym ignoring all modifiers */
         xcb_keysym_t keysym = keyresolv_get_keysym(ev->detail, 0);
         client_t *c;
-        if((c = client_getbyframewin(ev->event)))
+        if((c = client_getbywin(ev->event)))
         {
             luaA_object_push(globalconf.L, c);
             event_key_callback(ev, &c->keys, -1, 1, &keysym);
