@@ -103,6 +103,8 @@ typedef struct
     int keygrabber;
     /** The mouse pointer grabber function */
     int mousegrabber;
+    /** The drawable that currently contains the pointer */
+    drawable_t *drawable_under_mouse;
     /** Input focus information */
     struct
     {
@@ -110,6 +112,8 @@ typedef struct
         client_t *client;
         /** Is there a focus change pending? */
         bool need_update;
+        /** When nothing has the input focus, this window actually is focused */
+        xcb_window_t window_no_focus;
     } focus;
     /** Drawins */
     drawin_array_t drawins;
@@ -127,6 +131,8 @@ typedef struct
         bool registered;
         /** Systray window parent */
         drawin_t *parent;
+        /** Background color */
+        uint32_t background_pixel;
     } systray;
     /** The monitor of startup notifications */
     SnMonitorContext *snmonitor;
